/* BH2TEXT: Deompile a help file in the format used by Moonstone's BASHELP
 *
 *  Copyright (c) 2016 John Elliott <seasip.webmaster@gmail.com>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included 
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR 
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
 * OTHER DEALINGS IN THE SOFTWARE. */

/* This program has been written such that it will compile and run under
 * Hi-Tech C under CP/M. This means that it's designed to work in an 
 * environment with less than 64k of RAM; so any approach involving 
 * reading the whole file into memory is out. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#ifdef CPM		/* Standard C features not provided by Hi-Tech C */
#define const
#define EXIT_SUCCESS 0
#define EXIT_FAILURE 1
#define AV0 "BH2TEXT"
#else
#define AV0 argv[0]
#endif

FILE *st_fpin;	/* Input file (binary), must be rewindable */
FILE *st_fpout;	/* Output file (text) */

unsigned st_filepos = 0x80;	/* Used to deobfuscate input file */

/* In-memory representation of a topic header. */
typedef struct topic
{
	char *title;	/* Topic ID (malloced). If NULL, use tbuf. This is */
	char tbuf[10];	/* an optimisation preventing small mallocs */
	long offset;	/* Offset of topic in file */
} TOPIC;


TOPIC * st_topics;		/* List of loaded topics */
int st_topiccount;		/* Number of populated topics */
int st_topicmax;		/* Maximum number of topics */

char *indexbuf;			/* This will contain all topic IDs referenced
				 * by topic 0 (the index) */

/* In-memory representation of a hyperlink */
typedef struct hyperlink
{
	int row;		/* Location within page */
	int col;
	int width;		/* Number of characters in anchor */
	long offset;		/* Offset of target topic in file */
	char *text;		/* Displayed text */
} HYPERLINK;

/* In-memory representation of a page. This will be malloced to give 
 * space for 'links' hyperlinks. */
typedef struct pagedata
{
	char *text;		/* Page text (pointer to the buffer that
				 * was parsed to produce this page) */	
	int length;		/* Number of bytes in this page */
	int links;		/* Number of hyperlinks */
	HYPERLINK lk[1];	/* Hyperlinks (actual size given by 'links') */
} PAGEDATA;


/* Parse the raw binary representation of a page into a PAGEDATA structure */
PAGEDATA *parse_page(unsigned char *buf)
{
	PAGEDATA *pd;
	int len, n;
	int r, c;

	/* First byte is the count of hyperlinks, so we know how big the 
	 * structure will need to be */
	len = sizeof(PAGEDATA) + buf[0] * sizeof(HYPERLINK);
	pd = malloc(len);

	if (!pd)
	{
		fprintf(stderr, "parse_page: Could not allocate %d bytes.\n", len);
		exit(EXIT_FAILURE);
	}
	/* Populate link count */
	pd->links = buf[0];
	pd->length = 6 * pd->links + 1;
	/* Text begins after the hyperlinks */
	pd->text = (char *)(buf + (1 + 6 * pd->links));

	/* Populate each hyperlink */
	for (n = 0; n < pd->links; n++)
	{
		pd->lk[n].row    = r = buf[1 + 6*n] & 0xFF;
		pd->lk[n].col    = c = buf[2 + 6*n] & 0xFF;
		pd->lk[n].width  = buf[3 + 6*n] & 0xFF;
		pd->lk[n].offset = (buf[6 + 6*n] & 0xFF) +
			    128L * (buf[4 + 6*n] & 0xFF) +
			  32768L * (buf[5 + 6*n] & 0xFF);

		/* And locate the anchor text */
		pd->lk[n].text  = pd->text;
		while (r > 1)	
		{
			/* Skip over 'r' rows */
			while (pd->lk[n].text[0] != 0x0D &&
			       (pd->lk[n].text[0] & 0xFF) != 0xFF) 
			{
				++pd->lk[n].text;
			}
			/* (stopping if we hit the end of the page) */
			if ((pd->lk[n].text[0] & 0xFF) == 0xFF)
			{
				break;	
			}
			++pd->lk[n].text;
			--r;	
		}
		/* And skip over 'c' columns */
		while (c > 1)
		{
			if ((pd->lk[n].text[0] & 0xFF) == 0xFF)
			{
				break;	
			}
			++pd->lk[n].text;
			--c;	
		}
	}
	/* And now count the bytes in the page text to get total size */
	buf = (unsigned char *)(pd->text);
	while (*buf != 0xFF)
	{
		++buf;
		++pd->length;
	}
	++pd->length;
	return pd;
}



/* Rewind the input file; this requires the deobfuscator to be reinitialised. */
void rewind_file()
{
	st_filepos = 0x80;
	rewind(st_fpin);
}

/* Get the next byte from the input file, deobfuscating it */
int get_char()
{
	unsigned char a;
	int ch = fgetc(st_fpin);

	if (ch == EOF) return ch;

	a = 0x8F - ((ch ^ 0xA7) - st_filepos);
	++st_filepos;

	if (st_filepos == 0x100) st_filepos = 0x80;	
	return a;
}

/* Append a character to a malloced buffer, increasing the buffer size if 
 * the buffer is full. */
void append_char(char **pbuf, int *buflen, int *bufmax, int ch)
{
	(*pbuf)[*buflen] = ch;
	++(*buflen);
	/* The easy case: buffer is not full */
	if (*buflen < *bufmax)
	{
		return;
	}
	*pbuf = realloc(*pbuf, 1024 + *bufmax);
	if (!*pbuf)
	{
		fprintf(stderr, "append_char: could not allocate %d bytes.\n", 
				1024 + *bufmax);
		exit(EXIT_FAILURE);
	}
	*bufmax += 1024;
}

/* Load an encoded topic into a memory buffer. */
int load_topic(char **pbuf, int *plen)
{
	int ch;
	int pages, page, links, n;
	int buflen, bufmax;
	char *buf;

	/* Initialise the buffer with 1k */
	*pbuf = NULL;
	buf = malloc(1024);
	bufmax = 1024;
	buflen = 0;
	if (!buf)
	{
		fprintf(stderr, "load_topic: Could not allocate %d bytes.\n", bufmax);
		exit(EXIT_FAILURE);
	}

	/* Get count of pages */
	ch = get_char(); if (ch == EOF) { free(buf); return EOF; }
	append_char(&buf, &buflen, &bufmax, pages = ch);

	/* For each page: */
	for (page = 0; page < pages; page++)
	{
		/* Count of hyperlinks */
		ch = get_char(); if (ch == EOF) { free(buf); return EOF; }
		append_char(&buf, &buflen, &bufmax, links = ch);
		/* Hyperlinks */
		for (n = 0; n < 6 * links; n++)
		{
			ch = get_char(); 
			if (ch == EOF) { free(buf); return EOF; }
			append_char(&buf, &buflen, &bufmax, ch);
		}
		/* Body text */
		do
		{
			ch = get_char(); 
			if (ch == EOF) { free(buf); return EOF; }
			append_char(&buf, &buflen, &bufmax, ch);
		} while ((ch & 0xFF) != 0xFF);
	}
	*pbuf = buf;
	*plen = buflen;
	return 0;
}

/* Encode a topic name: mask all characters but alphanumerics and $ 
 * 
 * Space becomes _
 * Other banned characters become @ 
 */
char *encode_tname(const char *s)
{
	static char buf[256];
	int n;

	strncpy(buf, s, 255);
	buf[255] = 0;

	for (n = strlen(buf) - 1; n >= 0; n--)
	{
		if (!isalnum(buf[n]) && buf[n] != '$')
		{
			if (isspace(buf[n])) buf[n] = '_';
			else		     buf[n] = '@';
		}
	}
	return buf;
}

/* Look up the title of a topic -- its 'title' member if set, otherwise 
 * its 'tbuf' */
char *topictitle(int topic)
{
	if (st_topics[topic].title) return st_topics[topic].title;
	return st_topics[topic].tbuf;
}

/* Output a topic in decoded form */
void dump_topic(int topic)
{
	char *buf;
	int buflen;
	int page, pages, ch, n, m;

	/* Load the encoded form */
	if (load_topic(&buf, &buflen) != EOF)
	{
		char *ptr;
		int row, col;
/* On a CP/M system compiled C is slow; display what the program is doing so
 * it doesn't look like a crash.
 */
#if CPM
		printf("Writing topic: %s             \r", topictitle(topic));
#endif
/* Write topic title */
		fprintf(st_fpout, "[Topic=%s]\n", 
				encode_tname(topictitle(topic)));

/* Then for each page... */
		pages = buf[0];
		ch    = 1;
		for (page = 0; page < pages; page++)
		{
/* Parse the raw page data */
			PAGEDATA *pd = parse_page((unsigned char *)(buf + ch));

			row = 1;
			col = 1;
			ptr = pd->text;
/* Write "[Page]" */
			fprintf(st_fpout, "[Page]\n");
			while ((ptr[0] & 0xFF) != 0xFF)
			{
/* Precede each line with "|" */
				if (col == 1)
				{
					fputc('|', st_fpout);
				}
/* If a hyperlink starts at this position, write "[" which will
 * precede the anchor text */
				for (n = 0; n < pd->links; n++)
				{
					if (pd->lk[n].row == row &&
					    pd->lk[n].col == col)
					{
						fputc('[', st_fpout);
						break;
					}
				}
/* Output the character, skipping newlines and escaping those which need to 
 * be escaped. */
				switch (ptr[0] & 0xFF)
				{
					case 0x0D: case 0x0A: break;
					case 0x09: fprintf(st_fpout, "\\t"); break;
					case 0x1B: fprintf(st_fpout, "\\e"); break;
					case '[': 
					case ']': 
					case '|': 
					case '\\': fputc('\\', st_fpout);
						   /* FALL THROUGH */
					default:
						if (isprint(*ptr))
							fputc(*ptr, st_fpout);
						else fprintf(st_fpout, 
							"\\0x%02x", *ptr);
				}
/* If a hyperlink ends at this position, append "|target_id]" */

				for (n = 0; n < pd->links; n++)
				{
					if (pd->lk[n].row == row &&
					    pd->lk[n].col + pd->lk[n].width-1 == col)
					{
						int found = 0;

/* Look up its offset in the topics table to get the topic ID for that offset
 * (there should be one) */
						for (m = 0; m < st_topiccount; m++)
						{
							if (st_topics[m].offset ==
							    pd->lk[n].offset)
							{
								fprintf(st_fpout, "|%s]", encode_tname(topictitle(m)));
								found = 1;
								break;
							}
						}
/* But if we didn't, generate a dummy topic ID */
						if (!found) fprintf(st_fpout, 
							"|T%08lx]", pd->lk[n].offset);
					}
				}
/* Newline: End the line we're printing */
				if (*ptr == 0x0D) 
				{
					fprintf(st_fpout, "\n");
					col = 1;
					++row;
				}
				else	++col;
				++ptr;
			}
			ch += pd->length;
			free(pd);	/* End of page */
		}
		free(buf);	/* End of output, free the topic */
	}
}


/* First pass over the input file: Count the number of topics and 
 * populate the 'topics' array */
void enum_topics()
{
	int ch, found;
	int pages, page, n, m;
	char *buf;
	int buflen;
	long pos;
	int indexlen = 0;
	char *indexptr;
	PAGEDATA *pd;

	/* malloc topics in increments of 10 */
	st_topics = malloc(5 * sizeof(TOPIC));
	st_topicmax = 5;
	st_topiccount = 0;
	if (!st_topics)
	{
		fprintf(stderr, "enum_topics: Could not allocate %d bytes.\n",
				(int)(st_topicmax * sizeof(TOPIC)));
		exit(EXIT_FAILURE);
	}
	for (n = 0; n < st_topicmax; n++)
	{
		st_topics[n].title = NULL;
		st_topics[n].tbuf[0] = 0;
		st_topics[n].offset = 0;
	}
	rewind_file();

	pos = ftell(st_fpin);
#if CPM
	printf("Loading index\r");
#endif
	/* Load each topic */
	while (load_topic(&buf, &buflen) != EOF)
	{
		/* Grow st_topics if necessary */
		if (st_topiccount == st_topicmax)
		{
			st_topics = realloc(st_topics, 
				 (st_topicmax + 10) * sizeof(TOPIC));

			if (!st_topics)
			{
				fprintf(stderr, "enum_topics: Could not allocate %d bytes.\n",
				  (int)((st_topicmax + 10) * sizeof(TOPIC)));
				exit(EXIT_FAILURE);
			}
			st_topicmax += 10;
			for (n = st_topiccount; n < st_topicmax; n++)
			{
				st_topics[n].title = NULL;
			}
		}
/* Give topic a default ID: " Index" for the index, "Tnnnnnnnn" for others */
		if (!st_topiccount)
		{
			strcpy(st_topics[st_topiccount].tbuf, " Index");
		}
		else
		{
			sprintf(st_topics[st_topiccount].tbuf, "T%08lx", pos);
		}
		st_topics[st_topiccount].offset = pos;
		pos = ftell(st_fpin);
		++st_topiccount;
#if CPM
		printf("Topic %d        \r", st_topiccount); fflush(stdout);
#endif
		free(buf);
	}

	/* Now we have a list of topics, reload topic 0 (the table of 
	 * contents) and use it to generate topic IDs for all the other
	 * topics */
	rewind_file();
#if CPM
	printf("%-50.50s\r", "Reloading index");
#endif
	if (load_topic(&buf, &buflen) != EOF)
	{
		pages = buf[0];
		ch    = 1;

	/* Firstly, work out the total space that will be needed to hold
	 * all the topic IDs */

		for (page = 0; page < pages; page++)
		{
			pd = parse_page((unsigned char *)(buf + ch));

			for (n = 0; n < pd->links; n++)
			{
				indexlen += pd->lk[n].width + 1;
			}
			free(pd);
		}
	/* Then malloc that space */
		indexptr = malloc(indexlen);
		if (!indexptr)
		{
			fprintf(stderr, "enum_topics: Could not allocate %d bytes.\n",
				indexlen);
		}
	/* And then parse each page and extract the topic IDs */
		ch    = 1;
		for (page = 0; page < pages; page++)
		{
			pd = parse_page((unsigned char *)(buf + ch));

			for (n = 0; n < pd->links; n++)
			{
				memcpy(indexptr, pd->lk[n].text,
					pd->lk[n].width);
				indexptr[pd->lk[n].width] = 0;
				found = 0;
	/* As we encouter each link, search the topics table for 
	 * topics with that offset and give them an ID based on
	 * the link's anchor text. Linear search, I'm afraid. */
				for (m = 1; m < st_topiccount; m++)
				{
					if (st_topics[m].offset ==
						pd->lk[n].offset)
					{
						st_topics[m].title = indexptr;
						found = 1;
						break;
					}
				}
				if (!found)
				{
					fprintf(stderr, "Warning: Index entry '%s' not matched.\n", indexptr);
				}
				indexptr += 1 + pd->lk[n].width;
			}
			ch += pd->length;
			free(pd);
		}
		free(buf);	
	}
}

int main(int argc, char **argv)
{
	int n;

	if (argc < 2)
	{
		fprintf(stderr, "Syntax: %s helpfile {outfile}\n", AV0);
		return EXIT_FAILURE;
	}

	/* Open up the files. */
	st_fpin = fopen(argv[1], "rb");
	if (!st_fpin)
	{
		perror(argv[1]);
		return EXIT_FAILURE;
	}
	if (argc > 2)
	{
		st_fpout = fopen(argv[2], "wt");
		if (!st_fpout)
		{
			fclose(st_fpin);
			perror(argv[2]);
			return EXIT_FAILURE;

		}
	}
	else /* If no output file is given, use stdout */
	{
		st_fpout = stdout;
	}
#if CPM	/* This is slow on an 8-bit box, let's show some progress feedback */
	printf("%-50.50s\n", "Enumerating topics");
#endif
	/* Enumerate the topics */
	enum_topics();

	rewind_file();
#if CPM
	printf("%-50.50s\n", "Writing topics");
#endif

	/* And then decode each one */
	for (n = 0; n < st_topiccount; n++)
	{
		dump_topic(n);
	}
#if CPM
	printf("%-50.50s\n", "Complete");
#endif
	if (st_fpout != stdout && fclose(st_fpout))
	{
		fprintf(stderr, "Error closing output file\n");
	}
	return EXIT_SUCCESS;
}






