/* TEXT2BH: Compile a help file for Moonstone's BASHELP
 *
 *  Copyright (c) 2016 John Elliott <seasip.webmaster@gmail.com>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included 
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR 
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
 * OTHER DEALINGS IN THE SOFTWARE. */

/* This program has been written such that it will compile and run under
 * Hi-Tech C under CP/M. This means that it's designed to work in an 
 * environment with less than 64k of RAM; so any approach involving 
 * reading the whole file into memory is out. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#ifdef CPM		/* Standard C features not provided by Hi-Tech C */
#define const
#define EXIT_SUCCESS 0
#define EXIT_FAILURE 1
#define AV0 "TEXT2BH"
#else
#define AV0 argv[0]
#endif


FILE *st_fpin;		/* Input file (text, must be seekable) */
FILE *st_fpout;		/* Output file (binary) */

unsigned st_filepos = 0x80;	/* Used to obfuscate output file */

/* This represents a parsed hyperlink */
typedef struct hyperlink
{
	int row;	/* Location on the page of the hyperlink */
	int col;
	int width;	/* Width of the hyperlink */
	long offset;	/* Destination offset within the file */
	char *targetid;	/* Text ID of target topic, malloced */
} HYPERLINK;

/* This represents a parsed page, containing a block of text and possibly a
 * number of hyperlinks */
typedef struct page
{
	int textmax;	/* Max size of text buffer */
	int textlen;	/* Used size of text buffer */
	char *text;	/* Text buffer, malloced */
	int linkmax;	/* Max number of hyperlinks */
	int links;	/* Current number of hyperlinks */
	HYPERLINK *lk;	/* Hyperlinks, malloced */
} PAGE;

/* This represents a parsed topic */
typedef struct topic
{
	unsigned char pages;	/* Number of pages in topic */
	long offset;		/* Offset of topic in output file */
	long source;		/* Source of topic in input file */
	long length;		/* Length of encoded topic in output file */
	char *title;		/* Title of topic, malloced */
	PAGE *pg;		/* Array of pages, malloced */
} TOPIC;

TOPIC * st_topics;		/* List of loaded topics */
int st_topiccount;		/* Number of populated topics */
int st_topicmax;		/* Maximum number of topics */

/* Free memory structures associated with a PAGE */
void unload_page(PAGE *p)
{
	int n;

	/* For each hyperlink, free its target ID */
	for (n = 0; n < p->links; n++)
	{
		if (p->lk[n].targetid) free(p->lk[n].targetid);
	}
	/* Then free the text and hyperlink blocks */
	if (p->text) free(p->text);
	if (p->lk)   free(p->lk);

	p->text = NULL;
	p->lk   = NULL;
}

/* Free the page structures for a topic. But not the title; once that's 
 * loaded, it stays loaded. */
void unload_topic(TOPIC *t)
{
	int n;

	if (t->pg) 
	{
		/* Unload each page */
		for (n = 0; n < t->pages; n++)
		{
			unload_page(&t->pg[n]);	
		}
		/* Then free the pages array */
		free(t->pg);
		t->pg = NULL;
	}
}


/* Add a character to the text block of a page */
void append_char(PAGE *pg, char ch)
{
	/* If there's no room for it, */
	if (pg->textlen + 1 >= pg->textmax)
	{
		/* then allocate 1024 more bytes */
		pg->text = realloc(pg->text, pg->textmax + 1024);
		if (!pg->text)
		{
			fprintf(stderr, "append_char: Could not allocate %d "
				"bytes.\n", pg->textmax + 1024);
			exit(EXIT_FAILURE);
		}
		pg->textmax += 1024;
	}
	/* Append the character (and a terminating 0 just in case we 
	 * want to use strlen on it) */
	pg->text[pg->textlen++] = ch;
	pg->text[pg->textlen] = 0;
}

/* Add a hyperlink to the page's hyperlink collection */
void append_link(PAGE *pg, HYPERLINK *lk)
{
	/* If there's no room for it, */
	if (pg->links >= pg->linkmax)
	{
		/* then allocate 6 more hyperlinks */
		int size = (pg->linkmax + 6) * sizeof(HYPERLINK);
		pg->lk = realloc(pg->lk, size);
		if (!pg->lk)
		{
			fprintf(stderr, "append_link: Could not allocate %d "
				"bytes.\n", size);
			exit(EXIT_FAILURE);
		}
		pg->linkmax += 6;
	}
	/* Copy to the next free entry in the links array */
	memcpy(&pg->lk[pg->links], lk, sizeof(HYPERLINK));
	++pg->links;
}




/* Output a character to the target file, performing the record-based 
 * obfuscation used by BASHELP */
void put_char(unsigned char ch)
{
	unsigned char a;

	a = ((0x8f - ch) + st_filepos) ^ 0xa7;
	++st_filepos;
	if (st_filepos >= 0x100) st_filepos = 0x80;

	if (fputc(a, st_fpout) == EOF)
	{
		fprintf(stderr, "Write error on output file.\n");
		exit(EXIT_FAILURE);
	}
}


/* Encode a topic name: mask all characters but alphanumerics and $ 
 * 
 * Space becomes _
 * Other banned characters become @ 
 */
char *encode_tname(const char *s)
{
	static char buf[256];
	int n;

	strncpy(buf, s, 255);
	buf[255] = 0;

	for (n = strlen(buf) - 1; n >= 0; n--)
	{
		if (!isalnum(buf[n]) && buf[n] != '$')
		{
			if (isspace(buf[n])) buf[n] = '_';
			else		     buf[n] = '@';
		}
	}
	return buf;
}

/* Parse a \-preceded literal. This should be one of:
 * "\0xnn" -- hex literal
 * "\t"    -- tab
 * "\e"    -- escape
 * "\\     -- \
 * "\|     -- |
 * "\[     -- [
 * "\]     -- ]
 * 
 * Returns the address of the next character after the literal
 */
const char *parse_literal(const char *line, PAGE *pg)
{
	unsigned int res = 0;
	char xbuf[3];

	if (line[1] == '0' && line[2] == 'x')
	{
		sprintf(xbuf, "%-2.2s", line + 3);	
		line += 5;	/* "\0xXX" */
		sscanf(xbuf, "%x", &res);
	}
	else
	{
		res = line[1];
		if (res == 'e') res = 0x1B;
		if (res == 't') res = 0x09;
		line += 2;	/* "\c" */
	}
	append_char(pg, res);
	return line;
}

/* Parse a hyperlink, of the form "[text|target_id]" 
 * 
 * Returns the address of the next character after the hyperlink, and 
 * increases *col appropriately.
 */
const char *parse_hyperlink(const char *line, PAGE *pg, int row, int *col)
{
	int state = 0;	/* State is 0 if parsing "text", 1 for "target_id" */
	int tidpos = 0;	/* Length of target_id */
	HYPERLINK r;	/* This will be populated with the parsed hyperlink */
	char tidbuf[256]; /* Target ID that we parse out */

	r.row = row;	/* Row, column of hyperlink */
	r.col = *col;
	r.width = 0;
	r.offset = 0;	/* Initialise other bits of structure to 0 */
	r.targetid = NULL;

	++line;		/* Skip the opening '[' */
	while (line[0] != 0 && line[0] != ']')
	{
		/* The anchor text can contain \-literals, so parse them */
		if (state == 0 && line[0] == '\\')
		{
			line = parse_literal(line, pg);
			++r.width;
		}
		/* The pipe indicates end of anchor text */
		else if (state == 0 && line[0] == '|')
		{
			state = 1;
			line++;
		}
		/* Normal character in the anchor text */
		else if (state == 0) 
		{
			append_char(pg, line[0]);
			++r.width;
			line++;
		}
		/* Character in the target ID */
		else
		{
			if (tidpos < 256)
			{
				tidbuf[tidpos++] = line[0];
				tidbuf[tidpos] = 0;
			}
			line++;
		}
	}
	/* Skip over closing ] if there is one (there should be!) */
	if (line[0] == ']') ++line;

	/* Sanitise target name and add to the structure */
	strcpy(tidbuf, encode_tname(tidbuf));
	r.targetid = malloc(1 + strlen(tidbuf));
	if (!r.targetid)
	{
		fprintf(stderr, "parse_hyperlink: Could not allocate %d " 
			"bytes.\n", (int)(1 + strlen(tidbuf)));
		exit(EXIT_FAILURE);
	}
	strcpy(r.targetid, tidbuf);
	/* Link structure now populated, add it to the page */
	append_link(pg, &r);
	*col += r.width;
	return line;
}


/* Parse a line and append it to the specified page */
void load_line(const char *line, PAGE *pg, int row)
{
	int col = 1;
	while (*line)
	{
		switch(*line)
		{
			case '[':	/* Hyperlink */
				line = parse_hyperlink(line, pg, row, &col);
				break;

			case '\\': 	/* \-literal */
				line = parse_literal(line, pg);
				++col;
				break;
			default:	/* Normal character */
				append_char(pg, *line);
				++line;
				++col;
				break;
		}
	}
	/* Append a newline at the end of the line */
	append_char(pg, '\n');
}






/* Write a PAGE structure out in the BASHELP format. */
void encode_page(PAGE *pg)
{
	int n, m, found;

	put_char(pg->links);		/* 1 byte: Count of links */
	for (n = 0; n < pg->links; n++)
	{
		found = 0;
/* For each link, convert its target ID into an offset (linear search, I'm 
 * afraid) */
		for (m = 0; m < st_topiccount; m++)
		{
			if (!strcmp(st_topics[m].title, pg->lk[n].targetid))
			{
				pg->lk[n].offset = st_topics[m].offset;
				found = 1;
				break;
			}
		}
		if (!found)
		{
			fprintf(stderr, "Topic targetid '%s' not found\n",
					pg->lk[n].targetid);
		}
/* Encode the link as a 6-byte structure */
		put_char(pg->lk[n].row);
		put_char(pg->lk[n].col);
		put_char(pg->lk[n].width);
		put_char(pg->lk[n].offset / 128);
		put_char(pg->lk[n].offset / 32768);
		put_char(pg->lk[n].offset % 128);
	}
/* Now write out the page text. Newlines are encoded as 0x0D, all 
 * other characters as they are. */
	for (n = 0; n < pg->textlen; n++)
	{
		if (pg->text[n] == '\n') put_char(0x0D);
		else			 put_char(pg->text[n]);
	}
/* And zigamorph to indicate end of page text */
	put_char(0xff);
}



/* Load the pages for a topic into memory. Topic must already have its 
 * 'source' and 'pages' members populated */
void load_topic(TOPIC *t)
{
	int n, page, row;
	char *p, ltext[256];

	/* Seek to the topic */
	fseek(st_fpin, t->source, SEEK_SET);

#if CPM
	printf("Load topic: %s (%d page%s)         \r", t->title,
		t->pages, t->pages == 1 ? "" : "s"); fflush(stdout);
#endif

	/* Allocate the right number of pages */
	t->pg = malloc(sizeof(PAGE) * t->pages);
	if (!t->pg)
	{
		fprintf(stderr, "\nload_topic: Could not allocate %d bytes.\n",
				(int)(sizeof(PAGE) * t->pages));
		exit(EXIT_FAILURE);
	}
	/* Initialise all pages to blank */
	for (n = 0; n < t->pages; n++)
	{
		t->pg[n].textmax = 0;
		t->pg[n].textlen = 0;
		t->pg[n].linkmax = 0;
		t->pg[n].links = 0;
		t->pg[n].text = NULL;
		t->pg[n].lk = NULL;
	}
	page = 0;
	/* Then look for [Page] blocks, and parse them */
	while (page <= t->pages)
	{
		/* Load a line */
		if (!fgets(ltext, 255, st_fpin)) break;
		/* Trim off any newlines */
		p = strrchr(ltext, '\r'); if (p) *p = 0;
		p = strrchr(ltext, '\n'); if (p) *p = 0;
		/* Is it a [Page] line? */
		if (!strncmp(ltext, "[Page]", 6))
		{
			++page;
			row = 1;
			continue;
		}
		/* Or is it a text line? */
		if (ltext[0] == '|' && page > 0)
		{
			load_line(ltext + 1, &t->pg[page - 1], row);
			++row;
		}
	}
}


/* Scan the input file, determining the number of topics, how many pages they
 * have, and their titles */
void enum_topics()
{
	char *pch;
	long pos;
	int n;
	char ltext[256];

/* Start off with 5 topics, and grow in chunks of 5 */
	st_topics = malloc(5 * sizeof(TOPIC));
	st_topicmax = 5;
	st_topiccount = 0;
	if (!st_topics)
	{
		fprintf(stderr, "enum_topics: Could not allocate %d bytes.\n",
				(int)(st_topicmax * sizeof(TOPIC)));
		exit(EXIT_FAILURE);
	}
	for (n = 0; n < st_topicmax; n++)
	{
		st_topics[n].pages = 0;
		st_topics[n].offset = 0;
		st_topics[n].source = 0;
		st_topics[n].length = 0;
		st_topics[n].title = NULL;
		st_topics[n].pg = NULL;
	}
	rewind(st_fpin);

/* Parse the input file. Each line should be one of:
 *
 * "[Topic=topic_id]" 
 * "[Page]"
 * "|line_of_text"
 * 
 * At this point we're only interested in [Topic] and [Page] 
 */
	while (fgets(ltext, 255, st_fpin))
	{
		/* "[Page]" -- bump the page count on the last topic we
		 * saw (if there is one). [Page]s before [Topic]s are 
		 * ignored. */
		if (!strncmp(ltext, "[Page]", 6))
		{
			if (st_topiccount > 0)
			{
				++st_topics[st_topiccount - 1].pages;
			}
		}
		/* "[Topic=x]" */
		if (strncmp(ltext, "[Topic=", 7))
			continue;

		/* Chop off the closing bracket to leave the topic title */
		pch = strchr(ltext, ']');
		if (pch) *pch = 0;
/* On a CP/M system compiled C is slow; display what the program is doing so
 * it doesn't look like a crash.
 */
#if CPM
	printf("Topic: %s          \r", ltext + 7);
#endif
		pos = ftell(st_fpin);	
		/* If the topic array is full, reallocate it bigger */
		if (st_topiccount == st_topicmax)
		{
			st_topics = realloc(st_topics, 
				2 * st_topicmax * sizeof(TOPIC));

			if (!st_topics)
			{
				fprintf(stderr, "enum_topics: Could not allocate %d bytes\n",
				  (int)(2 * st_topicmax * sizeof(TOPIC)));
				exit(EXIT_FAILURE);
			}
			st_topicmax *= 2;
			/* And initialise the new elements */
			for (n = st_topiccount; n < st_topicmax; n++)
			{
				st_topics[n].pages = 0;
				st_topics[n].offset = 0;
				st_topics[n].source = 0;
				st_topics[n].length = 0;
				st_topics[n].title = NULL;
				st_topics[n].pg = NULL;
			}
		}
		st_topics[st_topiccount].source = pos;

/* In theory, we should be able to parse the topic titles and add them to
 * the structure at this point. In practice under CP/M, malloc()ing memory 
 * for the titles at the same time as we're realloc()ing memory for the topics 
 * array causes the memory arena to get badly fragmented and allocations to
 * fail. So we have to make two passes: one for the topics array, one for 
 * the titles.
 *
 * This is the code to do it in a single pass:
 */
/*		st_topics[n].title = malloc(1 + strlen(pch));
		if (!st_topics[n].title)
		{
			fprintf(stderr, "enum_topics: Could not allocate %d "
				 "bytes\n", (int)(1 + strlen(pch)));
			exit(EXIT_FAILURE);
		}
		strcpy(st_topics[n].title, pch);
 */

		++st_topiccount;
	}

	/* And this is the second pass to malloc the topic titles */
	rewind(st_fpin);
	n = 0;
	while (fgets(ltext, 255, st_fpin))
	{
		/* Look only for "[Topic=x]" lines */
		if (strncmp(ltext, "[Topic=", 7))
			continue;
		pch = strchr(ltext, ']');
		if (pch) *pch = 0;

#if CPM
		printf("Topic: %s          \r", ltext + 7);
#endif
		pch = encode_tname(ltext + 7);
		st_topics[n].title = malloc(1 + strlen(pch));
		if (!st_topics[n].title)
		{
			fprintf(stderr, "enum_topics: Could not allocate %d "
				 "bytes\n", (int)(1 + strlen(pch)));
			exit(EXIT_FAILURE);
		}
		strcpy(st_topics[n].title, pch);
		++n;
	}
}




/* Determine the encoded length of a topic */
void analyse_topic(int n)
{
	int page;
	int bytes = 1;

	/* Load the topic's pages */
	load_topic(&st_topics[n]);

	/* Size: 
	 *  1 byte link count
 	 *  6 bytes / link
	 *  length of page text
	 *  1 byte end marker */
	for (page = 0; page < st_topics[n].pages; page++)
	{
		bytes += 2 + 6 * st_topics[n].pg[page].links 
			   + st_topics[n].pg[page].textlen;
	}
	st_topics[n].length = bytes;
	/* And unload the pages again, memory is tight */
	unload_topic(&st_topics[n]);
}


/* Write out an encoded topic */
void encode_topic(int topic)
{
	int n;

	/* Which is just a matter of loading its pages, */
	load_topic(&st_topics[topic]);

	/* writing the count of pages, */
	put_char(st_topics[topic].pages);
	for (n = 0; n < st_topics[topic].pages; n++)
	{
		/* and writing each page in turn */
		encode_page(&st_topics[topic].pg[n]);
	}
	unload_topic(&st_topics[topic]);
}

int main(int argc, char **argv)
{
	int n;

	if (argc < 3)
	{
		fprintf(stderr, "Syntax: %s textfile helpfile\n", AV0);
		return EXIT_FAILURE;
	}

	/* Open up the files */
	st_fpin = fopen(argv[1], "rt");
	if (!st_fpin)
	{
		perror(argv[1]);
		return EXIT_FAILURE;
	}
	st_fpout = fopen(argv[2], "wb");
	if (!st_fpout)
	{
		fclose(st_fpin);
		perror(argv[2]);
		return EXIT_FAILURE;
	}
	/* And then get to work: */
#if CPM
	printf("%-50.50s\n", "Enumerating topics");
#endif
	/* 1. Count the topics and load their titles */
	enum_topics();
	st_topics[0].offset = 0;
#if CPM
	printf("%-50.50s\n", "Analysing topics");
#endif
	/* 2. Work out the sizes of all the topics, and hence their offsets */
	for (n = 0; n < st_topiccount; n++)
	{
		analyse_topic(n);
		if (n < (st_topiccount + 1))
		{
			st_topics[n+1].offset = st_topics[n].offset +
					        st_topics[n].length;
		}
	}
#if CPM
	printf("%-50.50s\n", "Encoding topics");
#endif
	/* 3. And now we have enough information to write the data out */
	for (n = 0; n < st_topiccount; n++)
	{
		encode_topic(n);
	}
	/* For strict compatibility we would need to pack the last record
	 * out to a multiple of 128 bytes with 0x1A 
	while (st_filepos & 0x7F) 
	{
		fputc(0x1A, st_fpout);
		++st_filepos;
	} */
#if CPM
	printf("%-50.50s\n", "Complete");
#endif
	if (fclose(st_fpout))
	{
		fprintf(stderr, "Error closing output file\n");
	}
	return EXIT_SUCCESS;
}






